__author__ = 'Jeffrey Knockel'
__copyright__ = 'Copyright 2013 Jeffrey Knockel'
__license__ = 'GPL v2'

from contextlib import closing
from copy import deepcopy
import os
import random
import socket
import time

from dpkt.ip import IP
from dpkt.ip6 import IP6, IP6FragmentHeader, ext_hdrs
from dpkt.icmp import ICMP, ICMP_ECHO
from dpkt.icmp6 import ICMP6, ICMP6_ECHO_REQUEST
from dpkt.tcp import TCP, TH_RST

from sniffing import tcp_string_to_flags

PID = os.getpid()
SO_BINDTODEVICE = 25

def default_src4(dst, iface=None):
    with closing(socket.socket(socket.AF_INET, socket.SOCK_RAW, socket.IPPROTO_RAW)) as s:
        if iface is not None:
            s.setsockopt(socket.SOL_SOCKET, SO_BINDTODEVICE, iface + '\0')
        s.connect((dst, 0))
        src = s.getsockname()[0]
    return src

def default_src6(dst, iface=None):
    with closing(socket.socket(socket.AF_INET6, socket.SOCK_RAW, socket.IPPROTO_RAW)) as s:
        if iface is not None:
            s.setsockopt(socket.SOL_SOCKET, SO_BINDTODEVICE, iface + '\0')
        s.connect((dst, 0))
        src = s.getsockname()[0]
    return src

def fill_data(seq, length):
    if length > 0:
        packed = '%.4x%.4x' % (PID, seq.data)
        seq.data = seq.data + 1 & 0xffff
        return packed * (length // 8) + packed[:length % 8]
    return ''

class MetaSequence(type):
    def __call__(cls, dst, src=None, iface=None, *args, **kwargs):
        dstn = cls.pton(dst)
        if src is None:
            src = cls.default_src(dst, iface)
        srcn = cls.pton(src)
        if cls.ADDR_PAIRS is None:
            cls.ADDR_PAIRS = {}
        pair = (dstn, srcn)
        if pair in cls.ADDR_PAIRS:
            return cls.ADDR_PAIRS[pair]
        new = cls.__new__(cls, dstn, srcn, *args, **kwargs)
        new.__init__(dstn, srcn, *args, **kwargs)
        cls.ADDR_PAIRS[pair] = new
        return new

class Sequence(object):
    __metaclass__ = MetaSequence
    ADDR_PAIRS = None

class IPSequence(Sequence):
    def __init__(self, ip_dst, ip_src, ip_id=None):
        if ip_id is None:
            ip_id = int(time.time())
        self.ip_dst = ip_dst
        self.ip_src = ip_src
        self.ip_id = ip_id
        self.sanitize_ip_id()
    @classmethod
    def ntop(cls, address):
        return socket.inet_ntop(socket.AF_INET, address)
    @classmethod
    def pton(cls, address):
        return socket.inet_pton(socket.AF_INET, address)
    @classmethod
    def default_src(cls, *args, **kwargs):
        return default_src4(*args, **kwargs)
    def next_ip_id(self):
        ip_id = self.ip_id
        self.ip_id += 1
        self.sanitize_ip_id()
        return ip_id
    def sanitize_ip_id(self):
        self.ip_id &= 0xffff
        if self.ip_id == 0:
            # with an AF_INET*/SOCK_RAW socket, the linux kernel will fill in
            # zero'd ip id's with its own
            self.ip_id = 1
    def next_packet_testing_mtu(self, mtu, **kwargs):
        return self.next_packet(mtu - 20, **kwargs)

class IP6Sequence(Sequence):
    def __init__(self, ip_dst, ip_src, ip_id=None):
        if ip_id is None:
            ip_id = int(time.time())
        self.ip_dst = ip_dst
        self.ip_src = ip_src
        self.ip_id = ip_id
        self.sanitize_ip_id()
    @classmethod
    def ntop(cls, address):
        return socket.inet_ntop(socket.AF_INET6, address)
    @classmethod
    def pton(cls, address):
        return socket.inet_pton(socket.AF_INET6, address)
    @classmethod
    def default_src(cls, *args, **kwargs):
        return default_src6(*args, **kwargs)
    def next_ip_id(self):
        ip_id = self.ip_id
        self.ip_id += 1
        self.sanitize_ip_id()
        return ip_id
    def sanitize_ip_id(self):
        self.ip_id &= 0xffffffff
        if self.ip_id == 0:
            # with an AF_INET*/SOCK_RAW socket, the linux kernel will fill in
            # zero'd ip id's with its own
            self.ip_id = 1
    def next_packet_testing_mtu(self, mtu, **kwargs):
        # mtu - 40, not mtu - 48, since we want to measure when the other host
        # begins including the fragmentation header in its response, despite
        # the fact that we always include it
        return self.next_packet(mtu - 40, **kwargs)

class ICMPSequence(IPSequence):
    def __init__(self, ip_dst, ip_src=None, ip_id=None, icmp_id=PID, icmp_seq=1, data=1):
        super(ICMPSequence, self).__init__(ip_dst, ip_src, ip_id)
        self.icmp_id = icmp_id
        self.icmp_seq = icmp_seq & 0xffff
        self.data = data & 0xffff
    def next_packet(self, data_length=64, type=ICMP_ECHO):
        data_length = max(data_length, 0)
        p = IP(len=data_length + 8, p=socket.IPPROTO_ICMP, id=self.next_ip_id(), dst=self.ip_dst, src=self.ip_src, data=
               ICMP(type=type, data=
               ICMP.Echo(id=self.icmp_id, seq=self.icmp_seq, data=
               fill_data(self, data_length))))
        self.icmp_seq = self.icmp_seq + 1 & 0xffff
        return p
    def next_packet_testing_mtu(self, mtu, *args, **kwargs):
        return super(ICMPSequence, self).next_packet_testing_mtu(mtu - 8, *args, **kwargs)

class ICMP6Sequence(IP6Sequence):
    def __init__(self, ip_dst, ip_src=None, ip_id=None, icmp_id=PID, icmp_seq=1, data=1):
        super(ICMP6Sequence, self).__init__(ip_dst, ip_src, ip_id)
        self.icmp_id = icmp_id
        self.icmp_seq = icmp_seq & 0xffff
        self.data = data & 0xffff
    def next_packet(self, data_length=64, type=ICMP6_ECHO_REQUEST):
        data_length = max(data_length, 0)
        p = IP6(hlim=64, plen=data_length + 16, nxt=44, dst=self.ip_dst, src=self.ip_src, data=
                ICMP6(type=type, data=
                ICMP6.Echo(id=self.icmp_id, seq=self.icmp_seq, data=
                fill_data(self, data_length))))
        p.p = socket.IPPROTO_ICMPV6
        p.extension_hdrs = dict(((i, None) for i in ext_hdrs))
        p.extension_hdrs[44] = IP6FragmentHeader(nxt=socket.IPPROTO_ICMPV6, id=self.next_ip_id())
        self.icmp_seq = self.icmp_seq + 1 & 0xffff
        return p
    def next_packet_testing_mtu(self, mtu, *args, **kwargs):
        return super(ICMP6Sequence, self).next_packet_testing_mtu(mtu - 8, *args, **kwargs)

class TCPSequenceMixin(object):
    def rst(self, p):
        rst = deepcopy(p)
        rst.id = self.next_ip_id()
        rst.data.flags = TH_RST
        rst.data.win = 0
        rst.data.seq = (rst.data.seq + 1) & 0xffffffff
        rst.data.ack = 0
        return rst
    def next_packet_with_rst(self, *args, **kwargs):
        p = self.next_packet(*args, **kwargs)
        return p, self.rst(p)

class TCPSequence(IPSequence, TCPSequenceMixin):
    def __init__(self, ip_dst, ip_src=None, ip_id=None, tcp_sport=None, tcp_dport=None, tcp_seq_random=None, tcp_ack_random=None, data=1):
        super(TCPSequence, self).__init__(ip_dst, ip_src, ip_id)
        if tcp_sport is None:
            tcp_sport = int(time.time())
        self.tcp_sport = (tcp_sport - 1024) % (32768 - 1024) + 1024
        if tcp_dport is None:
            tcp_dport = int(time.time()) + (32768 - 1024) // 2
        self.tcp_dport = (tcp_dport - 1024) % (32768 - 1024) + 1024
        if tcp_seq_random is None:
            tcp_seq_random = random.Random()
        self.tcp_seq_random = tcp_seq_random
        if tcp_ack_random is None:
            tcp_ack_random = random.Random()
        self.tcp_ack_random = tcp_ack_random
        self.data = data & 0xffff
    def next_packet(self, data_length=0, sport=None, dport=None, flags='S'):
        data_length = max(data_length, 0)
        if sport is None:
            sport = self.tcp_sport
            self.tcp_sport = (self.tcp_sport + 1 - 1024) % (32768 - 1024) + 1024
        if dport is None:
            dport = self.tcp_dport
            self.tcp_dport = (self.tcp_dport + 1 - 1024) % (32768 - 1024) + 1024
        ack = self.tcp_ack_random.randint(0, 0xffffffff) if 'A' in flags else 0
        p = IP(len=data_length + 8, p=socket.IPPROTO_TCP, id=self.next_ip_id(), dst=self.ip_dst, src=self.ip_src, data=
               TCP(sport=sport, dport=dport, seq=self.tcp_seq_random.randint(0, 0xffffffff), ack=ack, flags=tcp_string_to_flags(flags), win=8192, off_x2=(5 << 4), data=
               fill_data(self, data_length)))
        return p
    def next_packet_testing_mtu(self, mtu, *args, **kwargs):
        return super(TCPSequence, self).next_packet_testing_mtu(mtu - 20, *args, **kwargs)

class TCP6Sequence(IP6Sequence, TCPSequenceMixin):
    def __init__(self, ip_dst, ip_src=None, ip_id=None, tcp_sport=None, tcp_dport=None, tcp_seq_random=None, tcp_ack_random=None, data=1):
        super(TCP6Sequence, self).__init__(ip_dst, ip_src, ip_id)
        if tcp_sport is None:
            tcp_sport = int(time.time())
        self.tcp_sport = (tcp_sport - 1024) % (32768 - 1024) + 1024
        if tcp_dport is None:
            tcp_dport = int(time.time())
        self.tcp_dport = (tcp_dport - 1024) % (32768 - 1024) + 1024
        if tcp_seq_random is None:
            tcp_seq_random = random.Random()
        self.tcp_seq_random = tcp_seq_random
        if tcp_ack_random is None:
            tcp_ack_random = random.Random()
        self.tcp_ack_random = tcp_ack_random
        self.data = data & 0xffff
    def next_packet(self, data_length=0, sport=None, dport=None, flags='S'):
        data_length = max(data_length, 0)
        if sport is None:
            sport = self.tcp_sport
            self.tcp_sport = (self.tcp_sport + 1 - 1024) % (32768 - 1024) + 1024
        if dport is None:
            dport = self.tcp_dport
            self.tcp_dport = (self.tcp_dport + 1 - 1024) % (32768 - 1024) + 1024
        ack = self.tcp_ack_random.randint(0, 0xffffffff) if 'A' in flags else 0
        p = IP6(hlim=64, plen=data_length + 28, nxt=44, dst=self.ip_dst, src=self.ip_src, data=
                TCP(sport=sport, dport=dport, seq=self.tcp_seq_random.randint(0, 0xffffffff), ack=ack, flags=tcp_string_to_flags(flags), win=8192, off_x2=(5 << 4), data=
                fill_data(self, data_length)))
        p.p = socket.IPPROTO_TCP
        p.extension_hdrs = dict(((i, None) for i in ext_hdrs))
        p.extension_hdrs[44] = IP6FragmentHeader(nxt=socket.IPPROTO_TCP, id=self.next_ip_id())
        return p
    def next_packet_testing_mtu(self, mtu, *args, **kwargs):
        return super(TCP6Sequence, self).next_packet_testing_mtu(mtu - 20, *args, **kwargs)
