#!/usr/bin/env python

__author__ = 'Jeffrey Knockel'
__copyright__ = 'Copyright 2013 Jeffrey Knockel'
__license__ = 'GPL v2'

import sys
import time

from dpkt.tcp import TH_SYN, TH_ACK

from nslookup4 import lookup_or_die
from sequence import TCPSequence
from sniffing import tcp_flags_to_string, start_sniffing4 as start_sniffing
from tcp4 import receive_tcp

def synack_retransmit_times(sniffer, address, spoofed_addr, port, window=120.0, repeat=1, seconds=10, verbose=False):
    seq = TCPSequence(address, spoofed_addr)
    timeout = sniffer.timeout()
    requests = []
    sport_to_times = {}
    since = time.time() if verbose else None
    rsts = []
    def check_packet(p):
        start, times = sport_to_times[p.data.dport]
        delay = time.time() - start
        if delay < window:
            if p.data.flags & TH_SYN and p.data.flags & TH_ACK:
                times.append(delay)
            if verbose:
                now = time.time() - since
                sys.stderr.write('%f: %s on port %d with id %d after %f seconds\n' % (now, tcp_flags_to_string(p.data.flags), p.data.dport, p.id, delay))
    for _ in xrange(seconds):
        timeout.add_time(1.0)
        t = time.time()
        for _ in xrange(repeat):
            request, rst = seq.next_packet_with_rst(dport=port)
            rsts.append(rst)
            sniffer.send(request)
            requests.append(request)
            sport_to_times[request.data.sport] = t, []
        while True:
            p = receive_tcp(sniffer, requests, timeout)
            if p is None:
                break
            else:
                check_packet(p)
    timeout.add_time(window)
    while True:
        p = receive_tcp(sniffer, requests, timeout)
        if p is None:
            break
        else:
            check_packet(p)
    for rst in rsts:
        sniffer.send(rst)
    all_times = [sorted(times) for _, times in sport_to_times.itervalues()]
    max_len = max(len(times) for times in all_times)
    retransmit_times = [times for times in all_times if len(times) == max_len]
    n = len(retransmit_times)
    averaged = [sum(times) / n for times in zip(*retransmit_times)]
    avg_trans = float(sum(len(times) for times in all_times)) / (repeat * seconds)
    syn_frac = float(sum(len(times) > 0 for times in all_times)) / (repeat * seconds)
    return averaged, avg_trans, syn_frac

def usage(code=1):
    sys.stderr.write('Usage: %s HOST PORT SPOOFED_HOST\n' % sys.argv[0])
    exit(code)

def main():
    try:
        if len(sys.argv) != 4:
            usage()
        host = sys.argv[1]
        try:
            port = int(sys.argv[2])
        except ValueError:
            usage()
        spoofed_host = sys.argv[3]
        addr = lookup_or_die(host)[0]
        spoofed_addr = lookup_or_die(spoofed_host)[0]
        retransmit_times, avg_trans, syn_frac = start_sniffing(synack_retransmit_times, addr, spoofed_addr, port, verbose=True)
        print 'Average SYNACK\'s transmitted per SYN: %f' % avg_trans
        print 'Fraction of SYN\'s transmitting any SYNACK: %f' % syn_frac
        print 'Transmission times:'
        print retransmit_times
    except KeyboardInterrupt:
        sys.stderr.write('Received SIGINT\n')

if __name__ == '__main__':
    main()
