#!/usr/bin/env python

__author__ = 'Jeffrey Knockel'
__copyright__ = 'Copyright 2013 Jeffrey Knockel'
__license__ = 'GPL v2'

import sys
import time

from nslookup4 import lookup_or_die
from sequence import TCPSequence
from sniffing import start_sniffing4 as start_sniffing
from tcp4 import send_tcp, receive_tcp

def tcp_detect_ip_id_pattern(sniffer, addr, spoofed_addr, sport=None, dport=None, flags='S', threshold=5000):
    seq = TCPSequence(addr, iface=sniffer.iface)
    seq_spoofed = TCPSequence(addr, spoofed_addr)
    threshold += 1 # to account for our spoofed packet
    ip_id1 = None
    n = 0
    are_incrementing = True
    are_incrementing_by_one = True
    are_same = True
    are_echoed = True
    timeout = sniffer.timeout()
    min_time = float('inf')
    total_time = 0.0
    for _ in xrange(20):
        timeout.add_time(1.0)
        request = seq.next_packet(sport=sport, dport=dport, flags=flags)
        send_tcp(sniffer, request)
        start = time.time()
        p = receive_tcp(sniffer, request, timeout)
        ping_time = time.time() - start
        if p is not None:
            ip_id2 = p.id
            if ip_id1 is not None:
                min_time = min(min_time, ping_time)
                total_time += ping_time
                diff = (ip_id2 - ip_id1) & 0xffff
                are_incrementing = are_incrementing and 0 < diff <= threshold
                are_same = are_same and diff == 0
                are_incrementing_by_one = are_incrementing_by_one and diff == 1
                are_echoed = are_echoed and request.id == ip_id2
                n += 1
                if n >= 5:
                    break
            ip_id1 = ip_id2
            request = seq_spoofed.next_packet(sport=sport, dport=dport, flags=flags)
            send_tcp(sniffer, request)
        else:
            ip_id1 = None
        timeout.sleep()
    if n >= 5:
        if are_echoed:
            type = 'echoed'
        elif are_incrementing:
            if are_incrementing_by_one:
                type = 'per-host'
            else:
                type = 'globally'
        elif are_same:
            type = 'same'
        else:
            type = 'random'
    elif n > 0:
        type = 'unknown'
    else:
        type = 'none'
    avg_time = total_time / n if n > 0 else float('nan')
    return {'type': type, 'min_time': min_time, 'avg_time': avg_time}

def tcp_test_rate(sniffer, addr, repeat, seconds=10, sport=None, dport=None, flags='S'):
    seq = TCPSequence(addr, iface=sniffer.iface)
    timeout = sniffer.timeout()
    min_count = float('inf')
    max_count = 0
    total_count = 0
    for _ in xrange(seconds):
        timeout.add_time(1.0)
        sports = set()
        requests = []
        for _ in xrange(repeat):
            request = seq.next_packet(sport=sport, dport=dport, flags=flags)
            sports.add(request.data.sport)
            requests.append(request)
            send_tcp(sniffer, request)
        while True:
            p = receive_tcp(sniffer, requests, timeout)
            if p is None:
                break
            sports.discard(p.data.dport)
        count = repeat - len(sports)
        min_count = min(min_count, count)
        max_count = max(max_count, count)
        total_count += count
    return {'min': min_count, 'max': max_count, 'avg': float(total_count) / seconds}

def tests(sniffer, addr, spoofed_addr, port):
    print 'Testing %s...' % addr
    flags = 'SA'
    ip_id_pattern = tcp_detect_ip_id_pattern(sniffer, addr, spoofed_addr, sport=port, flags=flags)
    min_time = ip_id_pattern['min_time']
    avg_time = ip_id_pattern['avg_time']
    print 'TCP %s port %d IP id pattern type: %s' % (flags, port, ip_id_pattern['type'])
    print 'TCP %s port %d IP id minimum latency: %.4g ms' % (flags, port, round(min_time * 1000))
    print 'TCP %s port %d IP id average latency: %.4g ms' % (flags, port, round(avg_time * 1000))

def usage(code=1):
    sys.stderr.write('Usage: %s HOST PORT SPOOFED_HOST\n' % sys.argv[0])
    exit(code)

def main():
    try:
        if len(sys.argv) != 4:
            usage()
        host = sys.argv[1]
        try:
            port = int(sys.argv[2])
        except ValueError:
            usage()
        spoofed_host = sys.argv[3]
        addr = lookup_or_die(host)[0]
        spoofed_addr = lookup_or_die(spoofed_host)[0]
        start_sniffing(tests, addr, spoofed_addr, port)
    except KeyboardInterrupt:
        sys.stderr.write('Received SIGINT\n')

if __name__ == '__main__':
    main()
